# isort: skip_file
from .Patch import ImagePatches, VideoPatches


def __appropriate__(*args):
    """Says object was actually declared here, and not in the import module.
    Fixing sphinx warnings of not being able to find classes, when path is
    shortened.

    Parameters
    ----------
    *args
        The objects that you want sphinx to beleive that are defined here.

    Resolves `Sphinx referencing issues <https//github.com/sphinx-
    doc/sphinx/issues/3048>`
    """

    for obj in args:
        obj.__module__ = __name__


__appropriate__(
    ImagePatches,
    VideoPatches,
)
__all__ = [_ for _ in dir() if not _.startswith("_")]



import logging
import os

from functools import partial

from clapper.rc import UserDefaults
from sklearn.preprocessing import FunctionTransformer

import bob.io.base

from bob.bio.video import VideoLikeContainer
from bob.pad.base.database import FileListPadDatabase
from bob.pipelines import CSVToSamples, DelayedSample

logger = logging.getLogger(__name__)
rc = UserDefaults("bobrc.toml")


def load_multi_stream(path):
    data = bob.io.base.load(path)
    video = VideoLikeContainer(data[None, ...], [0])
    return video


def casia_surf_multistream_load(samples, original_directory):
    mod_to_attr = {}
    mod_to_attr["color"] = "filename"
    mod_to_attr["infrared"] = "ir_filename"
    mod_to_attr["depth"] = "depth_filename"
    mods = list(mod_to_attr.keys())

    def _load(sample):
        paths = dict()
        for mod in mods:
            paths[mod] = os.path.join(
                original_directory or "", getattr(sample, mod_to_attr[mod])
            )
        data = partial(load_multi_stream, paths["color"])
        depth = partial(load_multi_stream, paths["depth"])
        infrared = partial(load_multi_stream, paths["infrared"])
        subject = None
        key = sample.filename
        is_bonafide = sample.is_bonafide == "1"
        attack_type = None if is_bonafide else "attack"

        return DelayedSample(
            data,
            parent=sample,
            subject=subject,
            key=key,
            attack_type=attack_type,
            is_bonafide=is_bonafide,
            annotations=None,
            delayed_attributes={"depth": depth, "infrared": infrared},
        )

    return [_load(s) for s in samples]


def CasiaSurfMultiStreamSample(original_directory):
    return FunctionTransformer(
        casia_surf_multistream_load,
        kw_args=dict(original_directory=original_directory),
    )


class CasiaSurfPadDatabase(FileListPadDatabase):
    """The CASIA SURF Face PAD database interface.

    Parameters
    ----------
    stream_type : str
        A str or a list of str of the following choices: ``all``, ``color``, ``depth``, ``infrared``, by default ``all``

    The returned sample either have their data as a VideoLikeContainer or
    a dict of VideoLikeContainers depending on the chosen stream_type.
    """

    def __init__(
        self,
        **kwargs,
    ):
        original_directory = rc.get("bob.db.casia_surf.directory")
        if original_directory is None or not os.path.isdir(original_directory):
            raise FileNotFoundError(
                "The original_directory is not set. Please set it in the terminal using `bob config set bob.db.casia_surf.directory /path/to/database/CASIA-SURF/`."
            )
        transformer = CasiaSurfMultiStreamSample(
            original_directory=original_directory,
        )
        super().__init__(
            name="casia-surf",
            dataset_protocols_path=original_directory,
            protocol="all",
            reader_cls=partial(
                CSVToSamples,
                dict_reader_kwargs=dict(
                    delimiter=" ",
                    fieldnames=[
                        "filename",
                        "ir_filename",
                        "depth_filename",
                        "is_bonafide",
                    ],
                ),
            ),
            transformer=transformer,
            **kwargs,
        )
        self.annotation_type = None
        self.fixed_positions = None

    def protocols(self):
        return ["all"]

    def groups(self):
        return ["train", "dev", "eval"]

    def list_file(self, group):
        filename = {
            "train": "train_list.txt",
            "dev": "val_private_list.txt",
            "eval": "test_private_list.txt",
        }[group]
        return os.path.join(self.dataset_protocols_path, filename)



"""The `OULU-NPU`_ Database.
A mobile face presentation attack database with real-world variations database.

To configure the location of the database on your computer, run::

    bob config set bob.db.oulu_npu.directory /path/to/database/oulu-npu


If you use this database, please cite the following publication::

    @INPROCEEDINGS{OULU_NPU_2017,
             author = {Boulkenafet, Z. and Komulainen, J. and Li, Lei. and Feng, X. and Hadid, A.},
           keywords = {biometrics, face recognition, anti-spoofing, presentation attack, generalization, colour texture},
              month = May,
              title = {{OULU-NPU}: A mobile face presentation attack database with real-world variations},
            journal = {IEEE International Conference on Automatic Face and Gesture Recognition},
               year = {2017},
    }
"""
from bob.pad.face.database import OuluNpuPadDatabase

database = OuluNpuPadDatabase()



"""The `Mask-Attack`_ database for face anti-spoofing
consists of video clips of mask attacks. This database was produced at the
`Idiap Research Institute <http://www.idiap.ch>`_, in Switzerland.

If you use this database in your publication, please cite the following paper on
your references:


    @INPROCEEDINGS{ERDOGMUS_BTAS-2013,
        author = {Erdogmus, Nesli and Marcel, Sébastien},
        keywords = {biometric, Counter-Measures, Spoofing Attacks},
        month = september,
        title = {Spoofing in 2D Face Recognition with 3D Masks and Anti-spoofing with Kinect},
        journal = {Biometrics: Theory, Applications and Systems},
        year = {2013},}

After downloading, you can tell the bob library where the files are located
using::

    $ bob config set bob.db.mask_attack.directory /path/to/database/3dmad/Data/
"""
from bob.pad.face.database import MaskAttackPadDatabase

database = MaskAttackPadDatabase()



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Thu May 24 10:41:42 CEST 2012

from unittest import SkipTest

import numpy as np

import bob.bio.base


def test_replay_attack():
    database = bob.bio.base.load_resource(
        "replay-attack",
        "database",
        preferred_package="bob.pad.face",
        package_prefix="bob.pad.",
    )

    assert database.protocols() == [
        "digitalphoto",
        "grandtest",
        "highdef",
        "mobile",
        "photo",
        "print",
        "smalltest",
        "video",
    ]
    assert database.groups() == ["dev", "eval", "train"]
    assert len(database.samples(groups=["train", "dev", "eval"])) == 1200
    assert len(database.samples(groups=["train", "dev"])) == 720
    assert len(database.samples(groups=["train"])) == 360
    assert (
        len(database.samples(groups=["train", "dev", "eval"], purposes="real"))
        == 200
    )
    assert (
        len(
            database.samples(groups=["train", "dev", "eval"], purposes="attack")
        )
        == 1000
    )

    sample = database.sort(database.samples())[0]
    try:
        annot = dict(sample.annotations["0"])
        assert annot["leye"][1] > annot["reye"][1], annot
        assert annot == {
            "bottomright": [213, 219],
            "topleft": [58, 108],
            "leye": [118, 190],
            "reye": [117, 137],
            "mouthleft": [177, 144],
            "mouthright": [180, 183],
            "nose": [152, 164],
        }
        assert sample.data.shape == (20, 3, 240, 320)
        np.testing.assert_equal(sample.data[0][:, 0, 0], [8, 9, 11])
    except (RuntimeError, FileNotFoundError) as e:
        raise SkipTest(e)


def test_replay_mobile():
    database = bob.bio.base.load_resource(
        "replay-mobile",
        "database",
        preferred_package="bob.pad.face",
        package_prefix="bob.pad.",
    )

    assert database.protocols() == ["grandtest", "mattescreen", "print"]
    assert database.groups() == ["dev", "eval", "train"]
    assert len(database.samples(groups=["train", "dev", "eval"])) == 1030
    assert len(database.samples(groups=["train", "dev"])) == 728
    assert len(database.samples(groups=["train"])) == 312
    assert (
        len(database.samples(groups=["train", "dev", "eval"], purposes="real"))
        == 390
    )
    assert (
        len(
            database.samples(groups=["train", "dev", "eval"], purposes="attack")
        )
        == 640
    )

    all_samples = database.sort(database.samples())
    sample = all_samples[0]
    assert (
        sample.key
        == "devel/attack/attack_client005_session01_mattescreen_fixed_mobile_photo_lightoff.mov"
    ), sample.key
    assert sample.should_flip
    annot = dict(sample.annotations["0"])
    assert annot["leye"][1] > annot["reye"][1], annot
    assert annot == {
        "bottomright": [760, 498],
        "topleft": [374, 209],
        "leye": [518, 417],
        "reye": [522, 291],
        "mouthleft": [669, 308],
        "mouthright": [666, 407],
        "nose": [585, 358],
    }, annot

    sample2 = [s for s in all_samples if not s.should_flip][0]
    assert (
        sample2.key
        == "devel/attack/attack_client005_session01_mattescreen_fixed_tablet_photo_lightoff.mov"
    ), sample2.key
    assert not sample2.should_flip
    annot = dict(sample2.annotations["0"])
    assert annot["leye"][1] > annot["reye"][1], annot
    assert annot == {
        "reye": [873, 305],
        "leye": [879, 423],
        "nose": [937, 365],
        "mouthleft": [1018, 313],
        "mouthright": [1023, 405],
        "topleft": [747, 226],
        "bottomright": [1111, 495],
    }, annot

    try:
        assert sample.data.shape == (20, 3, 1280, 720), sample.data.shape
        np.testing.assert_equal(sample.data[0][:, 0, 0], [13, 13, 13])
        assert sample2.data.shape == (20, 3, 1280, 720), sample2.data.shape
        np.testing.assert_equal(sample2.data[0][:, 0, 0], [19, 33, 30])
    except (RuntimeError, FileNotFoundError) as e:
        raise SkipTest(e)


# Test the mask_attack database
def test_mask_attack():
    mask_attack = bob.bio.base.load_resource(
        "mask-attack",
        "database",
        preferred_package="bob.pad.face",
        package_prefix="bob.pad.",
    )
    # all real sequences: 2 sessions, 5 recordings for 17 individuals
    assert (
        len(
            mask_attack.samples(
                groups=["train", "dev", "eval"], purposes="real"
            )
        )
        == 170
    )
    # all attacks: 1 session, 5 recordings for 17 individuals
    assert (
        len(
            mask_attack.samples(
                groups=["train", "dev", "eval"], purposes="attack"
            )
        )
        == 85
    )

    # training real: 7 subjects, 2 sessions, 5 recordings
    assert len(mask_attack.samples(groups=["train"], purposes="real")) == 70
    # training real: 7 subjects, 1 session, 5 recordings
    assert len(mask_attack.samples(groups=["train"], purposes="attack")) == 35

    # dev and test contains the same number of sequences:
    # real: 5 subjects, 2 sessions, 5 recordings
    # attack: 5 subjects, 1 sessions, 5 recordings
    assert len(mask_attack.samples(groups=["dev"], purposes="real")) == 50
    assert len(mask_attack.samples(groups=["eval"], purposes="real")) == 50
    assert len(mask_attack.samples(groups=["dev"], purposes="attack")) == 25
    assert len(mask_attack.samples(groups=["eval"], purposes="attack")) == 25

    sample = mask_attack.samples()[0]
    try:
        assert sample.data.shape == (20, 3, 480, 640)
        np.testing.assert_equal(sample.data[0][:, 0, 0], [185, 166, 167])
        annot = sample.annotations["0"]
        assert annot["leye"][1] > annot["reye"][1], annot
        assert annot == {
            "leye": [212, 287],
            "reye": [217, 249],
        }
        assert sample.depth.shape == (20, 480, 640)
    except FileNotFoundError as e:
        raise SkipTest(e)


def test_casia_fasd():
    casia_fasd = bob.bio.base.load_resource(
        "casia-fasd",
        "database",
        preferred_package="bob.pad.face",
        package_prefix="bob.pad.",
    )

    assert len(casia_fasd.samples()) == 600
    assert len(casia_fasd.samples(purposes="real")) == 150
    assert len(casia_fasd.samples(purposes="attack")) == 450
    assert len(casia_fasd.samples(groups=("train", "dev"))) == 240
    assert len(casia_fasd.samples(groups="train")) == 180
    assert len(casia_fasd.samples(groups="dev")) == 60
    assert len(casia_fasd.samples(groups="eval")) == 360
    sample = casia_fasd.samples()[0]
    try:
        assert sample.data.shape == (20, 3, 480, 640)
        np.testing.assert_equal(sample.data[0][:, 0, 0], [217, 228, 227])
    except FileNotFoundError as e:
        raise SkipTest(e)


def test_casia_surf():
    try:
        casia_surf = bob.bio.base.load_resource(
            "casia-surf",
            "database",
            preferred_package="bob.pad.face",
            package_prefix="bob.pad.",
        )

        assert len(casia_surf.samples()) == 96584
        assert len(casia_surf.samples(purposes="real")) == 29394
        assert len(casia_surf.samples(purposes="attack")) == 67190
        assert len(casia_surf.samples(groups=("train", "dev"))) == 38874
        assert len(casia_surf.samples(groups="train")) == 29266
        assert len(casia_surf.samples(groups="dev")) == 9608
        assert len(casia_surf.samples(groups="eval")) == 57710
        sample = casia_surf.samples()[0]
        assert sample.data.shape == (1, 3, 279, 279)
        np.testing.assert_equal(sample.data[0][:, 0, 0], [0, 0, 0])
        assert sample.depth.shape == (1, 143, 143)
        assert sample.infrared.shape == (1, 143, 143)
    except FileNotFoundError as e:
        raise SkipTest(e)


def test_swan():
    database = bob.bio.base.load_resource(
        "swan",
        "database",
        preferred_package="bob.pad.face",
        package_prefix="bob.pad.",
    )

    assert database.protocols() == [
        "pad_p2_face_f1",
        "pad_p2_face_f2",
        "pad_p2_face_f3",
        "pad_p2_face_f4",
        "pad_p2_face_f5",
    ]
    assert database.groups() == ["dev", "eval", "train"]
    assert len(database.samples(groups=["train", "dev", "eval"])) == 5802
    assert len(database.samples(groups=["train", "dev"])) == 2803
    assert len(database.samples(groups=["train"])) == 2001
    assert (
        len(database.samples(groups=["train", "dev", "eval"], purposes="real"))
        == 3300
    )
    assert (
        len(
            database.samples(groups=["train", "dev", "eval"], purposes="attack")
        )
        == 2502
    )

    sample = database.sort(database.samples())[0]
    try:
        annot = dict(sample.annotations["0"])
        assert annot["leye"][1] > annot["reye"][1], annot
        assert annot == {
            "bottomright": [849, 564],
            "leye": [511, 453],
            "mouthleft": [709, 271],
            "mouthright": [711, 445],
            "nose": [590, 357],
            "reye": [510, 265],
            "topleft": [301, 169],
        }
        assert sample.data.shape == (20, 3, 1280, 720)
        np.testing.assert_equal(sample.data[0][:, 0, 0], [255, 255, 253])
    except (RuntimeError, FileNotFoundError) as e:
        raise SkipTest(e)


def test_oulu_npu():
    database = bob.bio.base.load_resource(
        "oulu-npu",
        "database",
        preferred_package="bob.pad.face",
        package_prefix="bob.pad.",
    )

    assert database.protocols() == [
        "Protocol_1",
        "Protocol_1_2",
        "Protocol_1_3",
        "Protocol_2",
        "Protocol_3_1",
        "Protocol_3_2",
        "Protocol_3_3",
        "Protocol_3_4",
        "Protocol_3_5",
        "Protocol_3_6",
        "Protocol_4_1",
        "Protocol_4_2",
        "Protocol_4_3",
        "Protocol_4_4",
        "Protocol_4_5",
        "Protocol_4_6",
    ]
    assert database.groups() == ["dev", "eval", "train"]
    assert (
        len(database.samples(groups=["train", "dev", "eval"]))
        == 1200 + 900 + 600
    )
    assert len(database.samples(groups=["train", "dev"])) == 1200 + 900
    assert len(database.samples(groups=["train"])) == 1200
    assert (
        len(database.samples(groups=["train", "dev", "eval"], purposes="real"))
        == 240 + 180 + 120
    )
    assert (
        len(
            database.samples(groups=["train", "dev", "eval"], purposes="attack")
        )
        == 960 + 720 + 480
    )

    sample = database.sort(database.samples())[0]
    try:
        annot = dict(sample.annotations["0"])
        assert annot["leye"][1] > annot["reye"][1], annot
        assert annot == {
            "bottomright": [1124, 773],
            "leye": [818, 638],
            "mouthleft": [1005, 489],
            "mouthright": [1000, 634],
            "nose": [906, 546],
            "reye": [821, 470],
            "topleft": [632, 394],
        }
        assert sample.data.shape == (20, 3, 1920, 1080)
        np.testing.assert_equal(sample.data[0][:, 0, 0], [195, 191, 199])
    except (RuntimeError, FileNotFoundError) as e:
        raise SkipTest(e)



import random

from collections import OrderedDict
from functools import partial

import numpy

from imageio import get_reader
from PIL import Image

import bob.io.image

from bob.bio.face.annotator import (
    bounding_box_from_annotation,
    min_face_size_validator,
)
from bob.bio.face.color import rgb_to_hsv, rgb_to_yuv
from bob.bio.video.annotator import normalize_annotations


def block(X, block_size, block_overlap, flat=False):
    """
    Parameters
    ----------
    X : numpy.ndarray
        The image to be split into blocks.
    block_size : tuple
        The size of the block.
    block_overlap : tuple
        The overlap of the block.

    Returns
    -------
    numpy.ndarray
        The image split into blocks.
    """

    assert len(block_size) == 2
    assert len(block_overlap) == 2

    size_ov_h = int(block_size[0] - block_overlap[0])
    size_ov_w = int(block_size[1] - block_overlap[1])
    n_blocks_h = int((X.shape[0] - block_overlap[0]) / size_ov_h)
    n_blocks_w = int((X.shape[1] - block_overlap[1]) / size_ov_w)

    blocks = numpy.zeros(shape=(n_blocks_h, n_blocks_w, size_ov_h, size_ov_w))
    for h in range(n_blocks_h):
        for w in range(n_blocks_w):
            blocks[h, w, :, :] = X[
                h * size_ov_h : h * size_ov_h + block_size[0],
                w * size_ov_w : w * size_ov_w + block_size[1],
            ]

    if flat:
        return blocks.reshape(
            n_blocks_h * n_blocks_w, blocks.shape[2], blocks.shape[3]
        )

    return blocks


def scale(image, scaling_factor):
    """
    Scales and image using PIL

    Parameters
    ----------

        image:
           Input image to be scaled

        new_shape: tuple
           Shape of the rescaled image



    """

    if isinstance(scaling_factor, float):
        new_size = tuple(
            (numpy.array(image.shape) * scaling_factor).astype(numpy.int)
        )
        return bob.io.image.to_bob(
            numpy.array(
                Image.fromarray(bob.io.image.to_matplotlib(image)).resize(
                    size=new_size
                ),
                dtype="float",
            )
        )

    elif isinstance(scaling_factor, tuple):
        if len(scaling_factor) > 2:
            scaling_factor = scaling_factor[1:]

        return bob.io.image.to_bob(
            numpy.array(
                Image.fromarray(bob.io.image.to_matplotlib(image)).resize(
                    size=scaling_factor
                ),
                dtype="float",
            )
        )
    else:
        raise ValueError(f"Scaling factor not supported: {scaling_factor}")


def frames(path):
    """Yields the frames of a video file.

    Parameters
    ----------
    path : str
        Path to the video file.

    Yields
    ------
    numpy.ndarray
        A frame of the video. The size is (3, 240, 320).
    """
    video = get_reader(path)
    for frame in video:
        yield bob.io.image.to_bob(frame)


def number_of_frames(path):
    """returns the number of frames of a video file.

    Parameters
    ----------
    path : str
        Path to the video file.

    Returns
    -------
    int
        The number of frames. Then, it yields the frames.
    """
    video = get_reader(path)
    return video.count_frames()


def bbx_cropper(frame, annotations):
    for source in ("direct", "eyes", None):
        try:
            bbx = bounding_box_from_annotation(source=source, **annotations)
            break
        except Exception:
            if source is None:
                raise
    return frame[..., bbx.top : bbx.bottom, bbx.left : bbx.right]


def min_face_size_normalizer(annotations, max_age=15, **kwargs):
    return normalize_annotations(
        annotations, partial(min_face_size_validator, **kwargs), max_age=max_age
    )


def yield_faces(pad_sample, cropper, normalizer=None):
    """Yields face images of a padfile. It uses the annotations from the
    database. The annotations are further normalized.

    Parameters
    ----------
    pad_sample
        The pad sample to return the faces.
    cropper : collections.abc.Callable
        A face image cropper that works with database's annotations.
    normalizer : collections.abc.Callable
        If not None, it should be a function that takes all the annotations of
        the whole video and yields normalized annotations frame by frame. It
        should yield same as ``annotations.items()``.

    Yields
    ------
    numpy.ndarray
        Face images

    Raises
    ------
    ValueError
        If the database returns None for annotations.
    """

    # read annotation
    annotations = pad_sample.annotations
    if annotations is None:
        raise ValueError("No annotations were returned.")

    if normalizer is not None:
        annotations = OrderedDict(normalizer(annotations))

    # normalize annotations and crop faces
    for frame_id, frame in enumerate(pad_sample.data):
        annot = annotations.get(str(frame_id), None)
        if annot is None:
            continue
        face = cropper(frame, annotations=annot)
        if face is not None:
            yield face


def scale_face(face, face_height, face_width=None):
    """Scales a face image to the given size.

    Parameters
    ----------
    face : numpy.ndarray
        The face image. It can be 2D or 3D in bob image format.
    face_height : int
        The height of the scaled face.
    face_width : :obj:`None`, optional
        The width of the scaled face. If None, face_height is used.

    Returns
    -------
    numpy.ndarray
        The scaled face.
    """
    face_width = face_height if face_width is None else face_width
    shape = list(face.shape)
    shape[-2:] = (face_height, face_width)
    # scaled_face = numpy.empty(shape, dtype="float64")
    scaled_face = scale(face, tuple(shape))
    return scaled_face


def blocks(data, block_size, block_overlap=(0, 0)):
    """Extracts patches of an image

    Parameters
    ----------
    data : numpy.ndarray
        The image in gray-scale, color, or color video format.
    block_size : (int, int)
        The size of patches
    block_overlap : (:obj:`int`, :obj:`int`), optional
        The size of overlap of patches

    Returns
    -------
    numpy.ndarray
        The patches.

    Raises
    ------
    ValueError
        If data dimension is not between 2 and 4 (inclusive).
    """

    data = numpy.asarray(data)
    # if a gray scale image:
    if data.ndim == 2:
        output = block(data, block_size, block_overlap, flat=True)
    # if a color image:
    elif data.ndim == 3:
        # out_shape = list(data.shape[0:1]) + list(
        #    block_output_shape(data[0], block_size, block_overlap, flat=True)
        # )

        # output = numpy.empty(out_shape, dtype=data.dtype)
        output = []
        for i, img2d in enumerate(data):
            output.append(block(img2d, block_size, block_overlap, flat=True))
        output = numpy.moveaxis(output, 0, 1)
    # if a color video:
    elif data.ndim == 4:
        output = [blocks(img3d, block_size, block_overlap) for img3d in data]
        output = numpy.concatenate(output, axis=0)
    else:
        raise ValueError("Unknown data dimension {}".format(data.ndim))
    return output


def block_generator(input, block_size, block_overlap=(0, 0)):
    """Performs a block decomposition of a 2D or 3D array/image

    It works exactly as :any:`bob.ip.base.block` except that it yields the blocks
    one by one instead of concatenating them. It also works with color images.

    Parameters
    ----------
    input : :any:`numpy.ndarray`
        A 2D array (Height, Width) or a color image (Bob format: Channels,
        Height, Width).
    block_size : (:obj:`int`, :obj:`int`)
        The size of the blocks in which the image is decomposed.
    block_overlap : (:obj:`int`, :obj:`int`), optional
        The overlap of the blocks.

    Yields
    ------
    array_like
        A block view of the image. Modifying the blocks will change the original
        image as well. This is different from :any:`bob.ip.base.block`.

    Raises
    ------
    ValueError
        If the block_overlap is not smaller than block_size.
        If the block_size is bigger than the image size.
    """
    block_h, block_w = block_size
    overlap_h, overlap_w = block_overlap
    img_h, img_w = input.shape[-2:]

    if overlap_h >= block_h or overlap_w >= block_w:
        raise ValueError(
            "block_overlap: {} must be smaller than block_size: {}.".format(
                block_overlap, block_size
            )
        )
    if img_h < block_h or img_w < block_w:
        raise ValueError(
            "block_size: {} must be smaller than the image size: {}.".format(
                block_size, input.shape[-2:]
            )
        )

    # Determine the number of block per row and column
    size_ov_h = block_h - overlap_h
    size_ov_w = block_w - overlap_w

    # Perform the block decomposition
    for h in range(0, img_h - block_h + 1, size_ov_h):
        for w in range(0, img_w - block_w + 1, size_ov_w):
            yield input[..., h : h + block_h, w : w + block_w]


def blocks_generator(data, block_size, block_overlap=(0, 0)):
    """Yields patches of an image

    Parameters
    ----------
    data : numpy.ndarray
        The image in gray-scale, color, or color video format.
    block_size : (int, int)
        The size of patches
    block_overlap : (:obj:`int`, :obj:`int`), optional
        The size of overlap of patches

    Yields
    ------
    numpy.ndarray
        The patches.

    Raises
    ------
    ValueError
        If data dimension is not between 2 and 4 (inclusive).
    """
    data = numpy.asarray(data)
    if 1 < data.ndim < 4:
        for patch in block_generator(data, block_size, block_overlap):
            yield patch
    # if a color video:
    elif data.ndim == 4:
        for frame in data:
            for patch in block_generator(frame, block_size, block_overlap):
                yield patch
    else:
        raise ValueError("Unknown data dimension {}".format(data.ndim))


def color_augmentation(image, channels=("rgb",)):
    """Converts an RGB image to different color channels.

    Parameters
    ----------
    image : numpy.ndarray
        The image in RGB Bob format.
    channels : :obj:`tuple`, optional
        List of channels to convert the image to. It can be any of ``rgb``,
        ``yuv``, ``hsv``.

    Returns
    -------
    numpy.ndarray
        The image that contains several channels:
        ``(3*len(channels), height, width)``.
    """
    final_image = []

    if "rgb" in channels:
        final_image.append(image)

    if "yuv" in channels:
        final_image.append(rgb_to_yuv(image))

    if "hsv" in channels:
        final_image.append(rgb_to_hsv(image))

    return numpy.concatenate(final_image, axis=0)


def random_sample(A, size):
    """Randomly selects ``size`` samples from the array ``A``"""
    return A[numpy.random.choice(A.shape[0], size, replace=False), ...]


def random_patches(image, block_size, n_random_patches=1):
    """Extracts N random patches of block_size from an image"""
    h, w = image.shape[-2:]
    bh, bw = block_size
    if h < block_size[0] or w < block_size[1]:
        raise ValueError("block_size must be smaller than image shape")
    hl = numpy.random.randint(0, h - bh, size=n_random_patches)
    wl = numpy.random.randint(0, w - bw, size=n_random_patches)
    for ch, cw in zip(hl, wl):
        yield image[..., ch : ch + bh, cw : cw + bw]


def extract_patches(
    image, block_size, block_overlap=(0, 0), n_random_patches=None
):
    """Yields either all patches from an image or N random patches."""
    if n_random_patches is None:
        return blocks_generator(image, block_size, block_overlap)
    else:
        return random_patches(
            image, block_size, n_random_patches=n_random_patches
        )


def the_giant_video_loader(
    pad_sample,
    region="whole",
    scaling_factor=None,
    cropper=None,
    normalizer=None,
    patches=False,
    block_size=(96, 96),
    block_overlap=(0, 0),
    random_patches_per_frame=None,
    augment=None,
    multiple_bonafide_patches=1,
    keep_pa_samples=None,
    keep_bf_samples=None,
):
    """Loads a video pad file frame by frame and optionally applies
    transformations.

    Parameters
    ----------
    pad_sample
        The pad sample
    region : str
        Either `whole` or `crop`. If whole, it will return the whole frame.
        Otherwise, you need to provide a cropper and a normalizer.
    scaling_factor : float
        If given, will scale images to this factor.
    cropper
        The cropper to use
    normalizer
        The normalizer to use
    patches : bool
        If true, will extract patches from images.
    block_size : tuple
        Size of the patches
    block_overlap : tuple
        Size of overlap of the patches
    random_patches_per_frame : int
        If not None, will only take this much patches per frame
    augment
        If given, frames will be transformed using this function.
    multiple_bonafide_patches : int
        Will use more random patches for bonafide samples
    keep_pa_samples : float
        If given, will drop some PA samples.
    keep_bf_samples : float
        If given, will drop some BF samples.

    Returns
    -------
    object
        A generator that yields the samples.

    Raises
    ------
    ValueError
        If region is not whole or crop.
    """
    if region == "whole":
        generator = iter(pad_sample.data)
    elif region == "crop":
        generator = yield_faces(
            pad_sample, cropper=cropper, normalizer=normalizer
        )
    else:
        raise ValueError("Invalid region value: `{}'".format(region))

    if scaling_factor is not None:
        generator = (scale(frame, scaling_factor) for frame in generator)
    if patches:
        if random_patches_per_frame is None:
            generator = (
                patch
                for frame in generator
                for patch in blocks_generator(frame, block_size, block_overlap)
            )
        else:
            if pad_sample.is_bonafide:
                random_patches_per_frame *= multiple_bonafide_patches
            generator = (
                patch
                for frame in generator
                for patch in random_sample(
                    blocks(frame, block_size, block_overlap),
                    random_patches_per_frame,
                )
            )

    if augment is not None:
        generator = (augment(frame) for frame in generator)

    if keep_pa_samples is not None and not pad_sample.is_bonafide:
        generator = (
            frame for frame in generator if random.random() < keep_pa_samples
        )

    if keep_bf_samples is not None and pad_sample.is_bonafide:
        generator = (
            frame for frame in generator if random.random() < keep_bf_samples
        )

    return generator



import bob.pipelines

from bob.bio.video import VideoLikeContainer
from bob.pad.face.transformer import VideoToFrames


def test_video_to_frames():
    # create a list of samples of videos
    # make sure some frames are None
    # call transform and check if the None frames are dropped

    videos = [[0, 1, 2, None, 3], [None, None, None]]
    video_container = [VideoLikeContainer(v, range(len(v))) for v in videos]
    samples = [
        bob.pipelines.Sample(v, key=i) for i, v in enumerate(video_container)
    ]
    frame_samples = VideoToFrames().transform(samples)
    assert len(frame_samples) == 4
    assert all(
        s.key == "0_{}".format(k) for s, k in zip(frame_samples, [0, 1, 2, 4])
    )
    assert all(s.video_key == 0 for s in frame_samples)
    assert [s.data for s in frame_samples] == [0, 1, 2, 3]
    assert [s.frame_id for s in frame_samples] == [0, 1, 2, 4]



import numpy
import pkg_resources

import bob.io.base as io
import bob.pipelines

from bob.bio.video import VideoLikeContainer


def _sample_video():
    path = pkg_resources.resource_filename(__name__, "data/test_image.png")
    img = io.load(path)
    video = VideoLikeContainer(img, [0])
    sample = bob.pipelines.Sample(video, key="sample", annotations=None)
    return sample


def test_pipeline():
    from bob.pad.face.config.deep_pix_bis import pipeline

    sample = _sample_video()
    prediction = pipeline.predict_proba([sample])[0]

    assert type(prediction.data) is numpy.float32

    assert prediction.data < 0.04



import os

import imageio
import numpy
import pkg_resources
import pytest

import bob.io.base

from bob.bio.video import VideoLikeContainer
from bob.pad.face.utils import (
    blocks,
    frames,
    number_of_frames,
    scale_face,
    yield_faces,
)
from bob.pipelines import DelayedSample


def get_pad_sample(none_annotations=False):
    def load():
        file_name = pkg_resources.resource_filename(
            __name__, "data/test_image_gs.png"
        )
        data = bob.io.base.load(file_name)[None, ...]
        indices = [os.path.basename(file_name)]
        fc = VideoLikeContainer(data, indices)
        return fc

    annotations = None
    if not none_annotations:
        annotations = {"0": {"topleft": (0, 0), "bottomright": (112, 92)}}

    return DelayedSample(
        load,
        client_id=123,
        key=0,
        attack_type=None,
        is_bonafide=True,
        annotations=annotations,
    )


image = get_pad_sample().data[0]


def test_video_frames():
    # get the path to cockatoo.mp4 from imageio-ffmpeg
    path = imageio.core.Request(
        "imageio:cockatoo.mp4", "r"
    ).get_local_filename()
    # read 2 frames
    for i, frame in enumerate(frames(path)):
        assert frame.shape == (3, 720, 1280), frame.shape
        assert frame.ndim == 3, frame.ndim
        if i == 0:
            numpy.testing.assert_equal(frame[:, 0, 0], [116, 119, 104])
        elif i == 1:
            numpy.testing.assert_equal(frame[:, 0, 0], [116, 119, 104])
        else:
            break

    # test number of frames
    n_frames = number_of_frames(path)
    assert n_frames == 280, n_frames


def dummy_cropper(frame, annotations=None):
    return frame


def test_yield_frames():
    padfile = get_pad_sample()
    nframes = len(padfile.data)
    assert nframes == 1, nframes
    for frame in padfile.data:
        assert frame.ndim == 2
        assert frame.shape == (312, 520)


def test_yield_faces_1():
    with pytest.raises(ValueError):
        padfile = get_pad_sample(none_annotations=True)
        for face in yield_faces(padfile, dummy_cropper):
            pass


def test_yield_faces_2():
    padfile = get_pad_sample(none_annotations=False)
    assert len(list(yield_faces(padfile, dummy_cropper)))
    for face in yield_faces(padfile, dummy_cropper):
        assert face.ndim == 2
        assert face.shape == padfile.data.shape[1:]


def test_scale_face():
    # gray-scale image
    face = image
    scaled_face = scale_face(face, 64)
    assert scaled_face.dtype == "float64"
    assert scaled_face.shape == (64, 64)
    # color image
    scaled_face = scale_face(numpy.array([face, face, face]), 64)
    assert scaled_face.dtype == "float64"
    assert scaled_face.shape == (3, 64, 64)
    assert (scaled_face[0] == scaled_face[1]).all()
    assert (scaled_face[0] == scaled_face[2]).all()


def test_blocks():
    # gray-scale image
    patches = blocks(image, (28, 28))
    assert patches.shape == (198, 28, 28), patches.shape
    # color image
    patches_gray = patches
    patches = blocks([image, image, image], (28, 28))
    assert patches.shape == (198, 3, 28, 28), patches.shape
    assert (patches_gray == patches[:, 0, ...]).all()
    assert (patches_gray == patches[:, 1, ...]).all()
    assert (patches_gray == patches[:, 2, ...]).all()
    # color video
    patches = blocks([[image, image, image]], (28, 28))
    assert patches.shape == (198, 3, 28, 28), patches.shape
    assert (patches_gray == patches[:, 0, ...]).all()
    assert (patches_gray == patches[:, 1, ...]).all()
    assert (patches_gray == patches[:, 2, ...]).all()


def test_block_raises1():
    with pytest.raises(ValueError):
        blocks(image[0], (28, 28))


def test_block_raises2():
    with pytest.raises(ValueError):
        blocks([[[image]]], (28, 28))



import dask_ml.model_selection as dcv

from sklearn.model_selection import StratifiedGroupKFold
from sklearn.pipeline import Pipeline
from sklearn.svm import SVC

from bob.bio.face.annotator import MTCNN
from bob.bio.face.preprocessor import INormLBP
from bob.bio.face.utils import make_cropper, pad_default_cropping
from bob.pad.face.transformer import VideoToFrames
from bob.pad.face.transformer.histogram import SpatialHistogram
from bob.pipelines.wrappers import SampleWrapper


def _init_pipeline(database, crop_size=(112, 112), grid_size=(3, 3)):
    # Face Crop
    # --------------------------
    annotator = MTCNN(thresholds=(0.1, 0.2, 0.2))
    cropped_pos = pad_default_cropping(crop_size, database.annotation_type)
    cropper = make_cropper(
        cropped_image_size=crop_size,
        cropped_positions=cropped_pos,
        fixed_positions=database.fixed_positions,
        color_channel="rgb",
        annotator=annotator,
    )
    face_cropper = SampleWrapper(
        cropper[0], transform_extra_arguments=cropper[1], delayed_output=False
    )

    # Extract LBP
    # --------------------------
    lbp_extractor = INormLBP(face_cropper=None, color_channel="gray")
    lbp_extractor = SampleWrapper(lbp_extractor, delayed_output=False)

    # Histogram
    # --------------------------
    histo = SpatialHistogram(grid_size=grid_size, nbins=256)
    # histo = VideoWrapper(histo)
    histo = SampleWrapper(histo, delayed_output=False)

    # Classifier
    # --------------------------
    sk_classifier = SVC()
    param_grid = [
        {
            "C": [2**p for p in range(-3, 14, 2)],
            "gamma": [2**p for p in range(-15, 0, 2)],
            "kernel": ["rbf"],
        }
    ]
    cv = StratifiedGroupKFold(n_splits=3)
    sk_classifier = dcv.GridSearchCV(
        sk_classifier, param_grid=param_grid, cv=cv
    )
    fit_extra_arguments = [("y", "is_bonafide"), ("groups", "video_key")]
    classifier = SampleWrapper(
        sk_classifier,
        delayed_output=False,
        fit_extra_arguments=fit_extra_arguments,
    )

    # Full Pipeline
    # --------------------------
    return Pipeline(
        [
            ("video2frames", VideoToFrames()),
            ("cropper", face_cropper),
            ("lbp", lbp_extractor),
            ("spatial_histogram", histo),
            ("classifier", classifier),
        ]
    )


# Get database information, needed for face cropper
db = globals()["database"]
if db is None:
    raise ValueError("Missing database!")
# Pipeline #
pipeline = _init_pipeline(database=db)



import logging

from functools import partial

from sklearn.base import BaseEstimator, TransformerMixin

from bob.pipelines.sample import DelayedSample, Sample
from bob.pipelines.wrappers import _frmt

logger = logging.getLogger(__name__)


def _get(sth):
    return sth


class VideoToFrames(TransformerMixin, BaseEstimator):
    """Expands video samples to frame-based samples only when transform is called."""

    def __init__(self, delayed_output=True):
        self.delayed_output = delayed_output

    def transform(self, video_samples):
        logger.debug(f"{_frmt(self)}.transform")
        outputs = []
        for vid_sample in video_samples:
            annotations = getattr(vid_sample, "annotations", None)
            # Define groups with `sample.key`` since we need a unique ID for
            # each video. The `groups` attribute is used to do cross-validation
            if not hasattr(vid_sample, "key"):
                raise ValueError(
                    "Video sample must have a unique `key` "
                    "attribute to be used with {}".format(
                        self.__class__.__name__
                    )
                )
            groups = vid_sample.key
            # video is an instance of VideoAsArray or VideoLikeContainer
            video = vid_sample.data
            for frame, frame_id in zip(video, video.indices):
                if frame is None:
                    continue
                # Do we have frame annotations?
                frame_annotations = None
                if annotations is not None:
                    # Global annotation are present -> query them
                    frame_annotations = annotations.get(str(frame_id))
                # Update key, otherwise get the one from parent and each frames
                # get the same one, breaking checkpoint mechanic for steps
                # later down the pipelines
                key = "{}_{}".format(vid_sample.key, frame_id)
                if self.delayed_output:
                    # create a load method so that we can create DelayedSamples
                    # because the input samples could be DelayedSamples with
                    # delayed attributes as well and we don't want to load
                    # those delayed attributes.
                    sample = DelayedSample(
                        partial(_get, frame),
                        frame_id=frame_id,
                        video_key=groups,
                        parent=vid_sample,
                        # Override parent's attributes
                        annotations=frame_annotations,
                        key=key,
                    )
                else:
                    sample = Sample(
                        frame,
                        frame_id=frame_id,
                        video_key=groups,
                        parent=vid_sample,
                        # Override parent's attributes
                        annotations=frame_annotations,
                        key=key,
                    )
                outputs.append(sample)
        return outputs

    def fit(self, X, y=None):
        return self

    def _more_tags(self):
        return {
            "requires_fit": False,
            "bob_checkpoint_features": False,
        }



"""The Swan_ Database.

To configure the location of the database on your computer, run::

    bob config set bob.db.swan.directory /path/to/database/swan


The Idiap part of the dataset comprises 150 subjects that are captured in six
different sessions reflecting real-life scenarios of smartphone assisted
authentication. One of the unique features of this dataset is that it is
collected in four different geographic locations representing a diverse
population and ethnicity. Additionally, it also contains a multimodal
Presentation Attack (PA) or spoofing dataset using low-cost Presentation Attack
Instruments (PAI) such as print and electronic display attacks. The novel
acquisition protocols and the diversity of the data subjects collected from
different geographic locations will allow developing a novel algorithm for
either unimodal or multimodal biometrics.

PAD protocols are created according to the SWAN-PAD-protocols document.
Bona-fide session 2 data is split into 3 sets of training, development, and
evaluation. The bona-fide data from sessions 3,4,5,6 are used for evaluation as
well. PA samples are randomly split into 3 sets of training, development, and
evaluation. All the random splits are done 10 times to created 10 different
protocols. The PAD protocols contain only one type of attacks. For convenience,
PA_F and PA_V protocols are created for face and voice, respectively which
contain all the attacks.
"""
from bob.pad.face.database import SwanPadDatabase

database = SwanPadDatabase()



"""`Replay-Mobile`_ is a database for face PAD experiments.

The Replay-Mobile Database for face spoofing consists of 1030 video clips of photo and video attack attempts to 40 clients,
under different lighting conditions.
These videos were recorded with current devices from the market -- an iPad Mini2 (running iOS) and a LG-G4 smartphone (running Android).
This Database was produced at the Idiap Research Institute (Switzerland) within the framework
of collaboration with Galician Research and Development Center in Advanced Telecommunications - Gradiant (Spain).
The reference citation is [CBVM16]_.

You can download the raw data of the `Replay-Mobile`_ database by following
the link. After downloading, you can tell the bob library where the files are
located using::

    $ bob config set bob.db.replay_mobile.directory /path/to/database/replay-mobile/database/
"""
from bob.pad.face.database import ReplayMobilePadDatabase

database = ReplayMobilePadDatabase()



import logging

from clapper.rc import UserDefaults

from bob.bio.base.database.utils import download_file
from bob.pad.base.database import FileListPadDatabase
from bob.pad.face.database import VideoPadSample

logger = logging.getLogger(__name__)
rc = UserDefaults("bobrc.toml")


def ReplayAttackPadDatabase(
    protocol="grandtest",
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
    annotation_directory=None,
    annotation_type=None,
    fixed_positions=None,
    **kwargs,
):
    name = "pad-face-replay-attack-aca6b46f.tar.gz"
    dataset_protocols_path = download_file(
        urls=[f"http://www.idiap.ch/software/bob/data/bob/bob.pad.face/{name}"],
        destination_filename=name,
        destination_sub_directory="protocols/pad",
        checksum="aca6b46f",
    )

    if annotation_directory is None:
        name = "annotations-replay-attack-mtcnn-8d1f4c12.tar.gz"
        annotation_directory = download_file(
            urls=[
                f"http://www.idiap.ch/software/bob/data/bob/bob.pad.face/{name}"
            ],
            destination_filename=name,
            destination_sub_directory="annotations",
            checksum="8d1f4c12",
        )
        annotation_type = "eyes-center"

    transformer = VideoPadSample(
        original_directory=rc.get("bob.db.replay_attack.directory"),
        annotation_directory=annotation_directory,
        selection_style=selection_style,
        max_number_of_frames=max_number_of_frames,
        step_size=step_size,
    )

    database = FileListPadDatabase(
        name="replay-attack",
        dataset_protocols_path=dataset_protocols_path,
        protocol=protocol,
        transformer=transformer,
        **kwargs,
    )
    database.annotation_type = annotation_type
    database.fixed_positions = fixed_positions
    return database



import logging
import os

from functools import partial

import h5py
import numpy as np

from clapper.rc import UserDefaults
from sklearn.preprocessing import FunctionTransformer

from bob.bio.base.database.utils import download_file
from bob.bio.video import VideoLikeContainer, select_frames
from bob.pad.base.database import FileListPadDatabase
from bob.pipelines import DelayedSample

logger = logging.getLogger(__name__)
rc = UserDefaults("bobrc.toml")


def load_frames_from_hdf5(
    hdf5_file,
    key="Color_Data",
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
):
    with h5py.File(hdf5_file) as f:
        video = f[key][()]
        # reduce the shape of depth from (N, C, H, W) to (N, H, W) since H == 1
        video = np.squeeze(video)

    indices = select_frames(
        len(video),
        max_number_of_frames=max_number_of_frames,
        selection_style=selection_style,
        step_size=step_size,
    )
    data = VideoLikeContainer(video[indices], indices)

    return data


def load_annotations_from_hdf5(
    hdf5_file,
):
    with h5py.File(hdf5_file) as f:
        eye_pos = f["Eye_Pos"][()]

    annotations = {
        str(i): {
            "reye": [row[1], row[0]],
            "leye": [row[3], row[2]],
        }
        for i, row in enumerate(eye_pos)
    }
    return annotations


def delayed_maskattack_video_load(
    samples,
    original_directory,
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
):
    original_directory = original_directory or ""
    results = []
    for sample in samples:
        hdf5_file = os.path.join(original_directory, sample.filename)
        data = partial(
            load_frames_from_hdf5,
            key="Color_Data",
            hdf5_file=hdf5_file,
            selection_style=selection_style,
            max_number_of_frames=max_number_of_frames,
            step_size=step_size,
        )
        depth = partial(
            load_frames_from_hdf5,
            key="Depth_Data",
            hdf5_file=hdf5_file,
            selection_style=selection_style,
            max_number_of_frames=max_number_of_frames,
            step_size=step_size,
        )
        annotations = partial(
            load_annotations_from_hdf5,
            hdf5_file=hdf5_file,
        )
        delayed_attributes = {
            "annotations": annotations,
            "depth": depth,
        }

        results.append(
            DelayedSample(
                data,
                parent=sample,
                delayed_attributes=delayed_attributes,
            )
        )
    return results


def MaskAttackPadSample(
    original_directory,
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
):
    return FunctionTransformer(
        delayed_maskattack_video_load,
        validate=False,
        kw_args=dict(
            original_directory=original_directory,
            selection_style=selection_style,
            max_number_of_frames=max_number_of_frames,
            step_size=step_size,
        ),
    )


def MaskAttackPadDatabase(
    protocol="classification",
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
    **kwargs,
):
    name = "pad-face-mask-attack-6d8854c2.tar.gz"
    dataset_protocols_path = download_file(
        urls=[f"http://www.idiap.ch/software/bob/data/bob/bob.pad.face/{name}"],
        destination_filename=name,
        destination_sub_directory="protocols/pad",
        checksum="6d8854c2",
    )

    transformer = MaskAttackPadSample(
        original_directory=rc.get("bob.db.mask_attack.directory"),
        selection_style=selection_style,
        max_number_of_frames=max_number_of_frames,
        step_size=step_size,
    )

    database = FileListPadDatabase(
        name="mask-attack",
        dataset_protocols_path=dataset_protocols_path,
        protocol=protocol,
        transformer=transformer,
        **kwargs,
    )
    database.annotation_type = "eyes-center"
    database.fixed_positions = None
    return database



import logging

import numpy as np
import torch
import torchvision.transforms as vision_transforms

from sklearn.base import BaseEstimator, ClassifierMixin
from torch import nn
from torchvision import models

from bob.bio.base.database.utils import download_file
from bob.io.image import to_matplotlib

logger = logging.getLogger(__name__)


DEEP_PIX_BIS_PRETRAINED_MODELS = {
    "oulu-npu-p1": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_1_model_0_0-24844429.pth"
    ],
    "oulu-npu-p2": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_2_model_0_0-4aae2f3a.pth"
    ],
    "oulu-npu-p3-1": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_3_1_model_0_0-f0e70cf3.pth"
    ],
    "oulu-npu-p3-2": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_3_2_model_0_0-92594797.pth"
    ],
    "oulu-npu-p3-3": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_3_3_model_0_0-71e18149.pth"
    ],
    "oulu-npu-p3-4": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_3_4_model_0_0-d7f666e5.pth"
    ],
    "oulu-npu-p3-5": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_3_5_model_0_0-fc40ba69.pth"
    ],
    "oulu-npu-p3-6": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_3_6_model_0_0-123a6c92.pth"
    ],
    "oulu-npu-p4-1": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_4_1_model_0_0-5f8dc7cf.pth"
    ],
    "oulu-npu-p4-2": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_4_2_model_0_0-168f2644.pth"
    ],
    "oulu-npu-p4-3": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_4_3_model_0_0-db57e3b5.pth"
    ],
    "oulu-npu-p4-4": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_4_4_model_0_0-e999b7e8.pth"
    ],
    "oulu-npu-p4-5": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_4_5_model_0_0-dcd13b8b.pth"
    ],
    "oulu-npu-p4-6": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_OULU_Protocol_4_6_model_0_0-96a1ab92.pth"
    ],
    "replay-mobile": [
        "http://www.idiap.ch/software/bob/data/bob/bob.pad.face/deep_pix_bis_RM_grandtest_model_0_0-6761ca7e.pth"
    ],
}
"A dictionary with the url paths to pre-trained weights of the DeepPixBis model."


class DeepPixBiS(nn.Module):
    """The class defining Deep Pixelwise Binary Supervision for Face Presentation
    Attack Detection:

    Reference: Anjith George and Sébastien Marcel. "Deep Pixel-wise Binary Supervision for
    Face Presentation Attack Detection." In 2019 International Conference on Biometrics (ICB).IEEE, 2019.

    Attributes
    ----------
    pretrained: bool
        If set to `True` uses the pretrained DenseNet model as the base. If set to `False`, the network
        will be trained from scratch.
    """

    def __init__(self, pretrained=True, **kwargs):
        """
        Parameters
        ----------
        pretrained: bool
            If set to `True` uses the pretrained densenet model as the base. Else, it uses the default network
        """
        super().__init__(**kwargs)

        dense = models.densenet161(pretrained=pretrained)

        features = list(dense.features.children())

        self.enc = nn.Sequential(*features[0:8])

        self.dec = nn.Conv2d(384, 1, kernel_size=1, padding=0)

        self.linear = nn.Linear(14 * 14, 1)

    def forward(self, x):
        """Propagate data through the network

        Parameters
        ----------
        img: :py:class:`torch.Tensor`
          The data to forward through the network. Expects RGB image of size 3x224x224

        Returns
        -------
        dec: :py:class:`torch.Tensor`
            Binary map of size 1x14x14
        op: :py:class:`torch.Tensor`
            Final binary score.

        """
        enc = self.enc(x)

        dec = self.dec(enc)

        dec = nn.Sigmoid()(dec)

        dec_flat = dec.view(-1, 14 * 14)

        op = self.linear(dec_flat)

        op = nn.Sigmoid()(op)

        return dec, op


class DeepPixBisClassifier(BaseEstimator, ClassifierMixin):
    """The class implementing the DeepPixBiS score computation"""

    def __init__(
        self,
        model_file=None,
        transforms=None,
        scoring_method="pixel_mean",
        device=None,
        threshold=0.8,
        **kwargs,
    ):
        """Init method

        Parameters
        ----------
        model_file: str
          The path of the trained PAD network to load or one of the keys to :py:attr:`DEEP_PIX_BIS_PRETRAINED_MODELS`
        transforms: :py:mod:`torchvision.transforms`
          Transform to be applied on the image
        scoring_method: str
          The scoring method to be used to get the final score,
          available methods are ['pixel_mean','binary','combined'].
        threshold: float
            The threshold to be used to binarize the output of the DeepPixBiS model.
            This is not used in the normal bob.pad.base pipeline.
        """
        super().__init__(**kwargs)

        if transforms is None:
            transforms = vision_transforms.Compose(
                [
                    vision_transforms.ToTensor(),
                    vision_transforms.Normalize(
                        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                    ),
                ]
            )

        # model
        self.transforms = transforms
        self.model = None
        self.scoring_method = scoring_method.lower()
        if self.scoring_method not in ("pixel_mean", "binary", "combined"):
            raise ValueError(
                "Scoring method {} is not implemented.".format(
                    self.scoring_method
                )
            )
        self.device = device
        self.threshold = threshold

        logger.debug(
            "Scoring method is : {}".format(self.scoring_method.upper())
        )

        if model_file in DEEP_PIX_BIS_PRETRAINED_MODELS:
            model_urls = DEEP_PIX_BIS_PRETRAINED_MODELS[model_file]
            filename = model_urls[0].split("/")[-1]
            file_hash = (
                model_urls[0].split("/")[-1].split("-")[-1].split(".")[0]
            )
            model_file = download_file(
                urls=model_urls,
                destination_filename=filename,
                destination_sub_directory="models",
                checksum=file_hash,
                extract=False,
            )

        logger.debug("Using pretrained model {}".format(model_file))
        self.model_file = model_file

    def load_model(self):
        if self.model is not None:
            return

        cp = torch.load(
            self.model_file, map_location=lambda storage, loc: storage
        )

        self.model = DeepPixBiS(pretrained=False)
        self.model.load_state_dict(cp["state_dict"])
        self.place_model_on_device()
        self.model.eval()
        logger.debug("Loaded the pretrained PAD model")

    def predict_proba(self, images):
        """Scores face images for PAD

        Parameters
        ----------
        image : 3D :py:class:`numpy.ndarray`
          The image to extract the score from. Its size must be 3x224x224;

        Returns
        -------
        output : float
          The output score is close to 1 for bonafide and 0 for PAs.
        """
        self.load_model()

        tensor_images = []
        for img in images:
            img = to_matplotlib(img)
            with torch.no_grad():
                img = self.transforms(img)
            tensor_images.append(img)

        images = tensor_images = torch.stack(tensor_images).to(self.device)

        with torch.no_grad():
            outputs = self.model.forward(images)

        output_pixel = outputs[0].cpu().detach().numpy().mean(axis=(1, 2, 3))
        output_binary = outputs[1].cpu().detach().numpy().mean(axis=1)

        score = {
            "pixel_mean": output_pixel,
            "binary": output_binary,
            "combined": (output_binary + output_pixel) / 2,
        }[self.scoring_method]

        print(score)
        return score

    def predict(self, X):
        scores = self.predict_proba(X)
        return np.int(scores > self.threshold)

    def fit(self, X, y=None):
        """No training required for this model"""
        return self

    def __getstate__(self):
        # Handling unpicklable objects
        d = self.__dict__.copy()
        d["model"] = None
        return d

    def _more_tags(self):
        return {"requires_fit": False}

    def place_model_on_device(self):
        if self.device is None:
            self.device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu"
            )
        if self.model is not None:
            self.model.to(self.device)



"""Config file for the CASIA FASD dataset.
Please run ``bob config set bob.db.casia_fasd.directory /path/to/database/casia_fasd/``
in terminal to point to the original files of the dataset on your computer."""

from bob.pad.face.database import CasiaFasdPadDatabase

database = CasiaFasdPadDatabase()



# isort: skip_file
from .load_utils import (  # noqa: F401
    frames,
    number_of_frames,
    yield_faces,
    scale_face,
    blocks,
    bbx_cropper,
    min_face_size_normalizer,
    color_augmentation,
    blocks_generator,
    the_giant_video_loader,
    random_sample,
    random_patches,
    extract_patches,
)

# gets sphinx autodoc done right - don't remove it
__all__ = [_ for _ in dir() if not _.startswith("_")]



"""`Replay-Attack`_ is a database for face PAD experiments.

The Replay-Attack Database for face spoofing consists of 1300 video clips of
photo and video attack attempts to 50 clients, under different lighting
conditions. This Database was produced at the Idiap Research Institute, in
Switzerland. The reference citation is [CAM12]_.

You can download the raw data of the `Replay-Attack`_ database by following the
link. After downloading, you can tell the bob library where the files are
located using::

    $ bob config set bob.db.replay_attack.directory /path/to/database/replay/protocols/replayattack-database/
"""
from bob.pad.face.database import ReplayAttackPadDatabase

database = ReplayAttackPadDatabase()



"""The `CASIA-SURF`_ database for face anti-spoofing

After downloading, you can tell the bob library where the files are located
using::

    $ bob config set bob.db.casia_surf.directory /path/to/database/CASIA-SURF/
"""
from bob.pad.face.database import CasiaSurfPadDatabase

database = CasiaSurfPadDatabase()



import logging

from clapper.rc import UserDefaults
from sklearn.pipeline import make_pipeline

from bob.bio.base.database.utils import download_file
from bob.pad.base.database import FileListPadDatabase
from bob.pad.face.database import VideoPadSample
from bob.pipelines.transformers import Str_To_Types, str_to_bool

logger = logging.getLogger(__name__)
rc = UserDefaults("bobrc.toml")


def get_rm_video_transform(sample):
    should_flip = sample.should_flip

    def transform(video):
        if not should_flip:
            # after changing to imageio-ffmpeg, we need to flip other way around
            video = video[..., ::-1]
        return video

    return transform


def ReplayMobilePadDatabase(
    protocol="grandtest",
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
    annotation_directory=None,
    annotation_type=None,
    fixed_positions=None,
    **kwargs,
):
    name = "pad-face-replay-mobile-620dded2.tar.gz"
    dataset_protocols_path = download_file(
        urls=[f"http://www.idiap.ch/software/bob/data/bob/bob.pad.face/{name}"],
        destination_filename=name,
        destination_sub_directory="protocols/pad",
        checksum="620dded2",
    )

    if annotation_directory is None:
        name = "annotations-replay-mobile-mtcnn-20055a07.tar.gz"
        annotation_directory = download_file(
            urls=[
                f"http://www.idiap.ch/software/bob/data/bob/bob.pad.face/{name}"
            ],
            destination_filename=name,
            destination_sub_directory="annotations",
            checksum="20055a07",
        )
        annotation_type = "eyes-center"

    transformer = make_pipeline(
        Str_To_Types(fieldtypes=dict(should_flip=str_to_bool)),
        VideoPadSample(
            original_directory=rc.get("bob.db.replay_mobile.directory"),
            annotation_directory=annotation_directory,
            selection_style=selection_style,
            max_number_of_frames=max_number_of_frames,
            step_size=step_size,
            get_transform=get_rm_video_transform,
        ),
    )

    database = FileListPadDatabase(
        name="replay-mobile",
        dataset_protocols_path=dataset_protocols_path,
        protocol=protocol,
        transformer=transformer,
        **kwargs,
    )
    database.annotation_type = annotation_type
    database.fixed_positions = fixed_positions
    return database



""" Deep Pixel-wise Binary Supervision for Face PAD

This baseline includes the models to replicate the experimental results published in the following publication::

    @INPROCEEDINGS{GeorgeICB2019,
        author = {Anjith George, Sebastien Marcel},
        title = {Deep Pixel-wise Binary Supervision for Face Presentation Attack Detection},
        year = {2019},
        booktitle = {ICB 2019},
    }

"""

from sklearn.pipeline import Pipeline

import bob.pipelines

from bob.bio.face.preprocessor import FaceCrop
from bob.bio.face.utils import get_default_cropped_positions
from bob.bio.video.transformer import VideoWrapper
from bob.pad.face.deep_pix_bis import DeepPixBisClassifier
from bob.pad.face.transformer import VideoToFrames

database = globals().get("database")
annotation_type, fixed_positions = None, None
if database is not None:
    annotation_type = database.annotation_type
    fixed_positions = database.fixed_positions

if annotation_type is None:
    annotation_type = "eyes-center"
    fixed_positions = None


# Preprocessor #
cropped_image_size = (224, 224)
cropped_positions = get_default_cropped_positions(
    "pad", cropped_image_size, annotation_type
)
cropper = FaceCrop(
    cropped_image_size=cropped_image_size,
    cropped_positions=cropped_positions,
    color_channel="rgb",
    fixed_positions=fixed_positions,
    dtype="uint8",
    annotator="mtcnn",
)

preprocessor = VideoWrapper(cropper)
preprocessor = bob.pipelines.wrap(
    ["sample"],
    preprocessor,
    transform_extra_arguments=(("annotations", "annotations"),),
)

# Classifier #
classifier = DeepPixBisClassifier(model_file="oulu-npu-p1")
classifier = bob.pipelines.wrap(["sample"], classifier)
# change the decision_function
decision_function = "predict_proba"


pipeline = Pipeline(
    [
        ("preprocessor", preprocessor),
        ("video_to_frames", VideoToFrames()),
        ("classifier", classifier),
    ]
)



"""Gets statistics on the average face size in a video database.
"""
import logging

from os.path import expanduser

import click
import numpy as np

from clapper.click import ConfigCommand, ResourceOption, verbosity_option

from bob.bio.face.annotator import (
    BoundingBox,
    bounding_box_from_annotation,
    expected_eye_positions,
)

logger = logging.getLogger(__name__)
BF = "Bona Fide"
PA = "Presentation Attack"


@click.command(entry_point_group="bob.bio.config", cls=ConfigCommand)
@click.option(
    "--database",
    "-d",
    required=True,
    cls=ResourceOption,
    entry_point_group="bob.pad.database",
)
@click.option("-o", "--output", default="face_sizes.png")
@click.option(
    "--database-directories-file",
    cls=ResourceOption,
    default=expanduser("~/.bob_bio_databases.txt"),
)
@verbosity_option(logger=logger, expose_value=False)
def statistics(database, output, database_directories_file, **kwargs):
    """Statistics on face size in video databases.

    \b
    Parameters
    ----------
    database : :any:`bob.pad.database`
        The database that you want to annotate. Can be a ``bob.pad.database``
        entry point or a path to a Python file which contains a variable
        named `database`.
    output : str
        Path to the saved figure.
    database_directories_file : str
        Path to a custom ``~/.bob_bio_databases.txt`` file.
    verbose : int, optional
        Increases verbosity (see help for --verbose).

    \b
    [CONFIG]...            Configuration files. It is possible to pass one or
                           several Python files (or names of ``bob.bio.config``
                           entry points) which contain the parameters listed
                           above as Python variables. The options through the
                           command-line (see below) will override the values of
                           configuration files.
    """
    logger.debug("database: %s", database)
    logger.debug("output: %s", output)
    logger.debug("database_directories_file: %s", database_directories_file)
    logger.debug("kwargs: %s", kwargs)

    # Some databases need their original_directory to be replaced
    database.replace_directories(database_directories_file)

    biofiles = database.objects(groups=None, protocol=database.protocol)
    biofiles = sorted(biofiles)

    logger.info("Gathering face size statistics of %d samples", len(biofiles))

    face_sizes_dict = {BF: [], PA: []}
    for i, biofile in enumerate(biofiles):
        for annot in database.annotations(biofile).values():
            # get the bounding box
            for source in ("direct", "eyes", None):
                try:
                    bbx = bounding_box_from_annotation(source=source, **annot)
                    break
                except Exception:
                    if source is None:
                        raise
                    else:
                        pass
            # record face size
            if biofile.attack_type is None:
                face_sizes_dict[BF].append(bbx.size)
            else:
                face_sizes_dict[PA].append(bbx.size)

    if output:
        import matplotlib.pyplot as plt

        # from matplotlib.backends.backend_pdf import PdfPages
        # pp = PdfPages(output)

    for attack_type, face_sizes in face_sizes_dict.items():
        click.echo(attack_type)
        face_sizes = np.array(face_sizes)
        # get statistics on the face sizes
        for name, array in (
            ("Height", face_sizes[:, 0]),
            ("Width", face_sizes[:, 1]),
        ):
            click.echo(
                "min: {}, mean: {}, max: {}, std: {:.1f} for {}".format(
                    array.min(),
                    int(array.mean()),
                    array.max(),
                    array.std(),
                    name,
                )
            )
        # print the average eye distance assuming bounding boxes are from
        # bob.ip.facedetect or the annotations had eye locations in them
        bbx = BoundingBox((0, 0), face_sizes.mean(axis=0))
        annot = expected_eye_positions(bbx)
        eye_distance = np.linalg.norm(
            np.array(annot["reye"]) - np.array(annot["leye"])
        )
        click.echo("Average eye locations: {}".format(annot))
        click.echo("Average eye distance: {}".format(int(eye_distance)))

        if not output:
            continue

        # plot the face sizes

        # plt.hist2d(face_sizes[:, 1], face_sizes[:, 0], bins=500)
        # plt.xlabel('Width')
        # plt.ylabel('Height')
        # plt.grid()

        # from matplotlib import cm
        # from mpl_toolkits.mplot3d import Axes3D
        # Z, xedges, yedges, _ = plt.hist2d(
        #     face_sizes[:, 1], face_sizes[:, 0], bins=500, normed=True)
        # xcenters = (xedges[:-1] + xedges[1:]) / 2
        # ycenters = (yedges[:-1] + yedges[1:]) / 2
        # X, Y = np.meshgrid(xcenters, ycenters)
        # fig = plt.figure()
        # ax = Axes3D(fig)
        # ax.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.viridis)

        plt.hist(
            face_sizes[:, 1],
            density=True,
            bins="auto",
            label=attack_type,
            alpha=0.5,
        )
    if output:
        plt.xlabel("Width of faces")
        plt.ylabel("Probability Density")

        plt.tight_layout()
        plt.legend()
        plt.savefig(output)
        # pp.savefig(plt.gcf())



import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array


def _get_cropping_size(image_size, patch_size):
    # How many pixels missing to cover the whole image
    r = image_size % patch_size
    # Spit gap into two evenly
    before = r // 2
    after = image_size - (r - before)
    return before, after


def _extract_patches(image, patch_size):
    # https://stackoverflow.com/a/16858283
    h, w = image.shape
    nrows, ncols = patch_size
    if h % nrows != 0 or w % ncols != 0:
        w_left, w_right = _get_cropping_size(w, ncols)
        h_top, h_bottom = _get_cropping_size(h, nrows)
        # Perform center crop
        image = image[h_top:h_bottom, w_left:w_right]
    return (
        image.reshape(h // nrows, nrows, -1, ncols)
        .swapaxes(1, 2)
        .reshape(-1, nrows, ncols)
    )


class SpatialHistogram(TransformerMixin, BaseEstimator):
    """
    Split images into a grid of patches, compute histogram on each one of them
    and concatenate them to obtain the final descriptor.
    """

    def __init__(self, grid_size=(4, 4), range=(0, 256), nbins=256):
        """
        Constructor
        :param grid_size: Tuple `(grid_y, grid_x)` indicating the number of
            patches to extract in each directions
        :param range: Tuple `(h_min, h_max)` indicating the histogram range.
            cf numpy.histogram
        :param nbins: Number of bins in the histogram, cf numpy.histogram
        """
        self.grid_size = grid_size
        self.range = range
        self.nbins = nbins

    def fit(self, X, y):
        return self

    def transform(self, X):
        X = check_array(X, allow_nd=True)  # X.shape == (N, H, W)
        histo = []
        for sample in X:
            h = self._spatial_histogram(sample)  # [grid_x * grid_y * nbins]
            histo.append(h)
        return np.asarray(histo)

    def _spatial_histogram(self, image):
        """Compute spatial histogram for a given images"""
        patch_size = [s // g for s, g in zip(image.shape, self.grid_size)]
        patches = _extract_patches(image=image, patch_size=patch_size)
        hist = []
        for patch in patches:
            h, _ = np.histogram(
                patch, bins=self.nbins, range=self.range, density=True
            )
            hist.append(h)
        return np.asarray(hist).reshape(-1)

    def _more_tags(self):
        return {"stateless": True, "requires_fit": False}



from collections import OrderedDict

from sklearn.base import BaseEstimator, TransformerMixin

from bob.bio.base.annotator.FailSafe import translate_kwargs
from bob.bio.video import VideoLikeContainer

from ..utils import extract_patches


class ImagePatches(TransformerMixin, BaseEstimator):
    """Extracts patches of images and returns it in a VideoLikeContainer. You need
    to wrap the further blocks (extractor and algorithm) that come after this
    in bob.bio.video wrappers.
    """

    def __init__(
        self, block_size, block_overlap=(0, 0), n_random_patches=None, **kwargs
    ):
        super(ImagePatches, self).__init__(**kwargs)
        self.block_size = block_size
        self.block_overlap = block_overlap
        self.n_random_patches = n_random_patches

    def transform(self, images):
        return [self.transform_one_image(img) for img in images]

    def transform_one_image(self, image):
        patches = extract_patches(
            image, self.block_size, self.block_overlap, self.n_random_patches
        )
        vc = VideoLikeContainer(patches, range(len(patches)))

        return vc


class VideoPatches(TransformerMixin, BaseEstimator):
    """Extracts patches of images from video containers and returns it in a
    VideoLikeContainer.
    """

    def __init__(
        self,
        face_cropper,
        block_size,
        block_overlap=(0, 0),
        n_random_patches=None,
        normalizer=None,
        **kwargs,
    ):
        super(VideoPatches, self).__init__(**kwargs)
        self.face_cropper = face_cropper
        self.block_size = block_size
        self.block_overlap = block_overlap
        self.n_random_patches = n_random_patches
        self.normalizer = normalizer

    def transform(self, videos, annotations=None):
        kwargs = translate_kwargs(dict(annotations=annotations), len(videos))
        return [
            self.transform_one_video(vid, **kw)
            for vid, kw in zip(videos, kwargs)
        ]

    def transform_one_video(self, frames, annotations=None):
        annotations = annotations or {}
        if self.normalizer is not None:
            annotations = OrderedDict(self.normalizer(annotations))

        all_patches = []
        for frame, index in zip(frames, frames.indices):
            # if annotations are given, and if particular frame annotations are
            # not missing we take them:
            annots = annotations.get(str(index))

            # preprocess image (by default: crops a face)
            preprocessed = self.face_cropper(frame, annots)
            if preprocessed is None:
                continue

            # extract patches
            patches = extract_patches(
                preprocessed,
                self.block_size,
                self.block_overlap,
                self.n_random_patches,
            )
            all_patches.extend(patches)

        vc = VideoLikeContainer(all_patches, range(len(all_patches)))

        if not len(vc):
            return None

        return vc



import os

from functools import partial

from sklearn.preprocessing import FunctionTransformer

from bob.bio.base.utils.annotations import read_annotation_file
from bob.bio.video import VideoAsArray
from bob.pipelines import DelayedSample


def get_no_transform(x):
    return None


def delayed_video_load(
    samples,
    original_directory,
    annotation_directory=None,
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
    get_transform=None,
    keep_extension_for_annotation=False,
):
    get_transform = get_transform or get_no_transform

    original_directory = original_directory or ""
    annotation_directory = annotation_directory or ""
    results = []
    for sample in samples:
        video_path = os.path.join(original_directory, sample.filename)
        data = partial(
            VideoAsArray,
            path=video_path,
            selection_style=selection_style,
            max_number_of_frames=max_number_of_frames,
            step_size=step_size,
            transform=get_transform(sample),
        )
        delayed_attributes = {"annotations": lambda: None}
        if annotation_directory:
            path = sample.filename
            if not keep_extension_for_annotation:
                path = os.path.splitext(sample.filename)[0]
            delayed_annotations = partial(
                read_annotation_file,
                file_name=f"{annotation_directory}:{path}.json",
                annotation_type="json",
            )
            delayed_attributes = {"annotations": delayed_annotations}
        if sample.attack_type == "":
            sample.attack_type = None
        sample.is_bonafide = sample.attack_type is None
        if not hasattr(sample, "key"):
            sample.key = sample.filename

        results.append(
            DelayedSample(
                data,
                parent=sample,
                delayed_attributes=delayed_attributes,
            )
        )
    return results


def VideoPadSample(
    original_directory,
    annotation_directory=None,
    selection_style=None,
    max_number_of_frames=None,
    step_size=None,
    get_transform=None,
    keep_extension_for_annotation=False,
):
    return FunctionTransformer(
        delayed_video_load,
        validate=False,
        kw_args=dict(
            original_directory=original_directory,
            annotation_directory=annotation_directory,
            selection_style=selection_style,
            max_number_of_frames=max_number_of_frames,
            step_size=step_size,
            get_transform=get_transform,
            keep_extension_for_annotation=keep_extension_for_annotation,
        ),
    )



# isort: skip_file
from .database import VideoPadSample
from .casia_fasd import CasiaFasdPadDatabase
from .casia_surf import CasiaSurfPadDatabase
from .mask_attack import MaskAttackPadDatabase
from .replay_attack import ReplayAttackPadDatabase
from .replay_mobile import ReplayMobilePadDatabase
from .swan import SwanPadDatabase
from .oulu_npu import OuluNpuPadDatabase


# gets sphinx autodoc done right - don't remove it
def __appropriate__(*args):
    """Says object was actually declared here, and not in the import module.
    Fixing sphinx warnings of not being able to find classes, when path is
    shortened. Parameters:

      *args: An iterable of objects to modify

    Resolves `Sphinx referencing issues
    <https://github.com/sphinx-doc/sphinx/issues/3048>`
    """

    for obj in args:
        obj.__module__ = __name__


__appropriate__(
    VideoPadSample,
    ReplayAttackPadDatabase,
    ReplayMobilePadDatabase,
    MaskAttackPadDatabase,
    CasiaSurfPadDatabase,
    CasiaFasdPadDatabase,
    SwanPadDatabase,
    OuluNpuPadDatabase,
)

__all__ = [_ for _ in dir() if not _.startswith("_")]
